==> CRTL-0.16.3-py3-none-any.whl: crtl/components/span_linking_scorer.py <==
from collections import Counter
from typing import Any, Callable, Dict, Iterable, Iterator, Literal, Optional, Tuple
from spacy.training import Example
from spacy.scorer import PRFScore
from spacy.util import registry
from spacy.tokens import Span


def overlap(span_1, span_2):
    x = range(span_1.start_char, span_1.end_char)
    y = range(span_2.start_char, span_2.end_char)
    xs = set(x)

    return len(xs.intersection(y)) > 0


def strong_annotation_match(predicted_span, reference_span):
    return (
        predicted_span.start_char == reference_span.start_char
        and predicted_span.end_char == reference_span.end_char
        and predicted_span.kb_id == reference_span.kb_id
    )


def weak_annotation_match(predicted_span, reference_span):
    return (
        overlap(predicted_span, reference_span)
        and predicted_span.kb_id == reference_span.kb_id
    )


class AbstractSpanSet:
    def __iter__(self) -> Iterator[Tuple[Iterator[Span], Iterator[Span]]]:
        raise NotImplementedError

    def _get_aligned_spans(self, doc, spans, align, allow_overlap):
        seen = set()
        output = []
        for span in spans:
            indices = align[span.start : span.end]
            if not allow_overlap:
                indices = [idx for idx in indices if idx not in seen]
            if len(indices) >= 1:
                aligned_span = Span(
                    doc,
                    indices[0],
                    indices[-1] + 1,
                    label=span.label,
                    kb_id=span.kb_id_,
                )
                target_text = span.text.lower().strip().replace(" ", "")
                our_text = aligned_span.text.lower().strip().replace(" ", "")
                if our_text == target_text:
                    output.append(aligned_span)
                    seen.update(indices)
        return output


class SpanSet(AbstractSpanSet):
    def __init__(self, examples, span_key):
        self._examples = examples
        self._span_key = span_key

    def __iter__(self) -> Iterator[Tuple[Iterator[Span], Iterator[Span]]]:
        for example in self._examples:
            prediction_spans = example.predicted.spans[self._span_key]
            reference_spans = self._get_aligned_spans(
                example.predicted,
                example.reference.spans[self._span_key],
                example.alignment.y2x,
                True,
            )
            yield prediction_spans, reference_spans


class EntSet(AbstractSpanSet):
    def __init__(self, examples):
        self._examples = examples

    def __iter__(self) -> Iterator[Tuple[Iterator[Span], Iterator[Span]]]:
        for example in self._examples:
            prediction_spans = example.predicted.ents
            reference_spans = self._get_aligned_spans(
                example.predicted, example.reference.ents, example.alignment.y2x, False
            )
            yield prediction_spans, reference_spans


@registry.scorers("crtl.score_links")
def make_linker_scorer(
    ent_or_span=Literal["ent", "span"], span_key: Optional[str] = None
) -> Callable:
    if ent_or_span == "ent":
        return get_ent_linker_scorer()
    elif ent_or_span == "span":
        if span_key:
            return get_span_linker_scorer(span_key)
        else:
            raise ValueError("span_key must be set if ent_or_span set to 'span'")
    else:
        raise ValueError("ent_or_span must be to 'ent' or 'span'")


def get_span_linker_scorer(span_key) -> Callable[[Iterable[Example]], Dict[str, Any]]:
    def span_linker_score(examples, **kwargs):
        return score_links(
            SpanSet(examples, span_key), negative_labels=["NIL"], **kwargs
        )

    return span_linker_score


def get_ent_linker_scorer():
    def ent_linker_score(examples, **kwargs):
        return score_links(EntSet(examples), negative_labels=["NIL"], **kwargs)

    return ent_linker_score


def score_links(
    spans: AbstractSpanSet, *, negative_labels: Iterator[str], **cfg
) -> Dict[str, Any]:
    # return {"strong_nel_ete_f":.8}
    return score_links_component_based(
        spans, negative_labels=negative_labels, **cfg
    ) | score_links_ete(spans, **cfg)


def score_links_ete(spans: AbstractSpanSet, **cfg):
    # see 'A Framework for Benchmarking Entity-Annotation Systems'

    score = {"STRONG": PRFScore(), "WEAK": PRFScore()}

    for strong_or_weak, match_predicate in [
        ("STRONG", strong_annotation_match),
        ("WEAK", weak_annotation_match),
    ]:
        for prediction_spans, reference_spans in spans:
            for predicted_span in prediction_spans:
                for reference_span in reference_spans:
                    if match_predicate(predicted_span, reference_span):
                        score[strong_or_weak].tp += 1
                        break
                else:
                    score[strong_or_weak].fp += 1
            for reference_span in reference_spans:
                for predicted_span in prediction_spans:
                    if match_predicate(predicted_span, reference_span):
                        break
                else:
                    score[strong_or_weak].fn += 1

    results = {
        f"weak_nel_ete_p": score["WEAK"].precision,
        f"weak_nel_ete_r": score["WEAK"].recall,
        f"weak_nel_ete_f": score["WEAK"].fscore,
        f"strong_nel_ete_p": score["STRONG"].precision,
        f"strong_nel_ete_r": score["STRONG"].recall,
        f"strong_nel_ete_f": score["STRONG"].fscore,
    }
    return results


def update_score_with_example(
    prediction_spans: Iterable[Span],
    reference_spans: Iterable[Span],
    f_per_type: Dict[str, Any],
    *,
    negative_labels: Iterable[str],
) -> None:
    gold_span_by_offset = {}
    for gold_span in reference_spans:
        gold_span_by_offset[(gold_span.start_char, gold_span.end_char)] = gold_span
    for pred_span in prediction_spans:
        gold_span = gold_span_by_offset.get(
            (pred_span.start_char, pred_span.end_char), None
        )
        if gold_span is not None:
            label = gold_span.label_
            if label not in f_per_type:
                f_per_type[label] = PRFScore()
            try:
                gold = gold_span.kb_id_
            except:
                gold = "NIL"
            # only evaluating entities that overlap between gold and pred,
            # to disentangle the performance of the NEL from the NER
            if gold is not None:
                pred = pred_span.kb_id_
                if gold in negative_labels and pred in negative_labels:
                    # ignore true negatives
                    pass
                elif gold == pred:
                    f_per_type[label].tp += 1
                elif gold in negative_labels:
                    f_per_type[label].fp += 1
                elif pred in negative_labels:
                    f_per_type[label].fn += 1
                else:
                    # a wrong prediction (e.g. Q42 != Q3) counts as both a FP as well as a FN
                    f_per_type[label].fp += 1
                    f_per_type[label].fn += 1


def score_links_component_based(
    spans: AbstractSpanSet, *, negative_labels: Iterable[str], **cfg
) -> Dict[str, Any]:
    """Returns PRF for predicted links on the entity level.
    To disentangle the performance of the NEL from the NER,
    this method only evaluates NEL links for entities that overlap
    between the gold reference and the predictions.
    examples (Iterable[Example]): Examples to score
    negative_labels (Iterable[str]): The string values that refer to no annotation (e.g. "NIL")
    RETURNS (Dict[str, Any]): A dictionary containing the scores.
    DOCS: https://spacy.io/api/scorer#score_links
    """
    # modified score_links from https://github.com/explosion/spaCy/blob/master/spacy/scorer.py
    f_per_type: Dict[str, Any] = {}
    for prediction_spans, reference_spans in spans:
        update_score_with_example(
            prediction_spans,
            reference_spans,
            f_per_type,
            negative_labels=negative_labels,
        )
    micro_prf = PRFScore()
    for label_prf in f_per_type.values():
        micro_prf.tp += label_prf.tp
        micro_prf.fn += label_prf.fn
        micro_prf.fp += label_prf.fp
    n_labels = len(f_per_type) + 1e-100
    macro_p = sum(prf.precision for prf in f_per_type.values()) / n_labels
    macro_r = sum(prf.recall for prf in f_per_type.values()) / n_labels
    macro_f = sum(prf.fscore for prf in f_per_type.values()) / n_labels
    results = {
        f"nel_score": micro_prf.fscore,
        f"nel_score_desc": "micro F",
        f"nel_micro_p": micro_prf.precision,
        f"nel_micro_r": micro_prf.recall,
        f"nel_micro_f": micro_prf.fscore,
        f"nel_macro_p": macro_p,
        f"nel_macro_r": macro_r,
        f"nel_macro_f": macro_f,
        f"nel_f_per_type": {k: v.to_dict() for k, v in f_per_type.items()},
    }
    return results

==> CRTL-0.22.0-py3-none-any.whl: crtl/components/span_linking_scorer.py <==
from collections import Counter
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
)
from spacy.training import Example
from spacy.scorer import PRFScore
from spacy.util import registry
from spacy.tokens import Span


def overlap(span_1, span_2):
    x = range(span_1.start_char, span_1.end_char)
    y = range(span_2.start_char, span_2.end_char)
    xs = set(x)

    return len(xs.intersection(y)) > 0


def strong_annotation_match(predicted_span, reference_span):
    return (
        predicted_span.start_char == reference_span.start_char
        and predicted_span.end_char == reference_span.end_char
        and predicted_span.kb_id == reference_span.kb_id
    )


def weak_annotation_match(predicted_span, reference_span):
    return (
        overlap(predicted_span, reference_span)
        and predicted_span.kb_id == reference_span.kb_id
    )


class AbstractSpanSet:
    def __iter__(self) -> Iterator[Tuple[Iterator[Span], Iterator[Span]]]:
        raise NotImplementedError

    def _get_aligned_spans(self, doc, spans, align, allow_overlap):
        seen = set()
        output = []
        for span in spans:
            indices = align[span.start : span.end]
            if not allow_overlap:
                indices = [idx for idx in indices if idx not in seen]
            if len(indices) >= 1:
                aligned_span = Span(
                    doc,
                    indices[0],
                    indices[-1] + 1,
                    label=span.label,
                    kb_id=span.kb_id_,
                )
                target_text = span.text.lower().strip().replace(" ", "")
                our_text = aligned_span.text.lower().strip().replace(" ", "")
                if our_text == target_text:
                    output.append(aligned_span)
                    seen.update(indices)
        return output


class SpanSet(AbstractSpanSet):
    def __init__(self, examples, span_key):
        self._examples = examples
        self._span_key = span_key

    def __iter__(self) -> Iterator[Tuple[Iterator[Span], Iterator[Span]]]:
        for example in self._examples:
            prediction_spans = example.predicted.spans[self._span_key]
            reference_spans = self._get_aligned_spans(
                example.predicted,
                example.reference.spans[self._span_key],
                example.alignment.y2x,
                True,
            )
            yield prediction_spans, reference_spans


class EntSet(AbstractSpanSet):
    def __init__(self, examples):
        self._examples = examples

    def __iter__(self) -> Iterator[Tuple[Iterator[Span], Iterator[Span]]]:
        for example in self._examples:
            prediction_spans = example.predicted.ents
            reference_spans = self._get_aligned_spans(
                example.predicted, example.reference.ents, example.alignment.y2x, False
            )
            yield prediction_spans, reference_spans


@registry.scorers("crtl.score_links")
def make_linker_scorer(
    ent_or_span=Literal["ent", "span"],
    span_key: Optional[str] = None,
    label_subset: Optional[List[str]] = None,
) -> Callable:
    if ent_or_span == "ent":
        return get_ent_linker_scorer(label_subset)
    elif ent_or_span == "span":
        if span_key:
            return get_span_linker_scorer(span_key, label_subset)
        else:
            raise ValueError("span_key must be set if ent_or_span set to 'span'")
    else:
        raise ValueError("ent_or_span must be to 'ent' or 'span'")


def get_span_linker_scorer(
    span_key, label_subset: Optional[List[str]] = None
) -> Callable[[Iterable[Example]], Dict[str, Any]]:
    def span_linker_score(examples, **kwargs):
        return score_links(
            SpanSet(examples, span_key),
            negative_labels=["NIL"],
            label_subset=label_subset,
            **kwargs,
        )

    return span_linker_score


def get_ent_linker_scorer(label_subset: Optional[List[str]] = None):
    def ent_linker_score(examples, **kwargs):
        return score_links(
            EntSet(examples),
            negative_labels=["NIL"],
            label_subset=label_subset,
            **kwargs,
        )

    return ent_linker_score


def score_links(
    spans: AbstractSpanSet,
    *,
    negative_labels: Iterator[str],
    label_subset: Optional[List[str]] = None,
    **cfg,
) -> Dict[str, Any]:
    # return {"strong_nel_ete_f":.8}
    return score_links_component_based(
        spans, negative_labels=negative_labels, label_subset=label_subset, **cfg
    ) | score_links_ete(spans, label_subset=label_subset, **cfg)


def score_links_ete(spans: AbstractSpanSet, label_subset: Optional[List[str]], **cfg):
    # see 'A Framework for Benchmarking Entity-Annotation Systems'

    score = {"STRONG": PRFScore(), "WEAK": PRFScore()}

    for strong_or_weak, match_predicate in [
        ("STRONG", strong_annotation_match),
        ("WEAK", weak_annotation_match),
    ]:
        for prediction_spans, reference_spans in spans:
            filtered_prediction_spans = [
                span
                for span in prediction_spans
                if not label_subset or span.label_ in label_subset
            ]
            filtered_reference_spans = [
                span
                for span in reference_spans
                if not label_subset or span.label_ in label_subset
            ]
            for predicted_span in filtered_prediction_spans:
                for reference_span in filtered_reference_spans:
                    if match_predicate(predicted_span, reference_span):
                        score[strong_or_weak].tp += 1
                        break
                else:
                    score[strong_or_weak].fp += 1
            for reference_span in filtered_reference_spans:
                for predicted_span in filtered_prediction_spans:
                    if match_predicate(predicted_span, reference_span):
                        break
                else:
                    score[strong_or_weak].fn += 1

    results = {
        f"weak_nel_ete_p": score["WEAK"].precision,
        f"weak_nel_ete_r": score["WEAK"].recall,
        f"weak_nel_ete_f": score["WEAK"].fscore,
        f"strong_nel_ete_p": score["STRONG"].precision,
        f"strong_nel_ete_r": score["STRONG"].recall,
        f"strong_nel_ete_f": score["STRONG"].fscore,
    }
    return results


def update_score_with_example(
    prediction_spans: Iterable[Span],
    reference_spans: Iterable[Span],
    f_per_type: Dict[str, Any],
    *,
    negative_labels: Iterable[str],
    label_subset: Optional[List[str]],
) -> None:
    gold_span_by_offset = {}
    filtered_prediction_spans = [
        span
        for span in prediction_spans
        if not label_subset or span.label_ in label_subset
    ]
    filtered_reference_spans = [
        span
        for span in reference_spans
        if not label_subset or span.label_ in label_subset
    ]
    for gold_span in filtered_reference_spans:
        gold_span_by_offset[(gold_span.start_char, gold_span.end_char)] = gold_span
    for pred_span in filtered_prediction_spans:
        gold_span = gold_span_by_offset.get(
            (pred_span.start_char, pred_span.end_char), None
        )
        if gold_span is not None:
            label = gold_span.label_
            if label not in f_per_type:
                f_per_type[label] = PRFScore()
            try:
                gold = gold_span.kb_id_
            except:
                gold = "NIL"
            # only evaluating entities that overlap between gold and pred,
            # to disentangle the performance of the NEL from the NER
            if gold is not None:
                pred = pred_span.kb_id_
                if gold in negative_labels and pred in negative_labels:
                    # ignore true negatives
                    pass
                elif gold == pred:
                    f_per_type[label].tp += 1
                elif gold in negative_labels:
                    f_per_type[label].fp += 1
                elif pred in negative_labels:
                    f_per_type[label].fn += 1
                else:
                    # a wrong prediction (e.g. Q42 != Q3) counts as both a FP as well as a FN
                    f_per_type[label].fp += 1
                    f_per_type[label].fn += 1


def score_links_component_based(
    spans: AbstractSpanSet,
    *,
    negative_labels: Iterable[str],
    label_subset: Optional[List[str]],
    **cfg,
) -> Dict[str, Any]:
    """Returns PRF for predicted links on the entity level.
    To disentangle the performance of the NEL from the NER,
    this method only evaluates NEL links for entities that overlap
    between the gold reference and the predictions.
    examples (Iterable[Example]): Examples to score
    negative_labels (Iterable[str]): The string values that refer to no annotation (e.g. "NIL")
    RETURNS (Dict[str, Any]): A dictionary containing the scores.
    DOCS: https://spacy.io/api/scorer#score_links
    """
    # modified score_links from https://github.com/explosion/spaCy/blob/master/spacy/scorer.py
    f_per_type: Dict[str, Any] = {}
    for prediction_spans, reference_spans in spans:
        update_score_with_example(
            prediction_spans,
            reference_spans,
            f_per_type,
            negative_labels=negative_labels,
            label_subset=label_subset,
        )
    micro_prf = PRFScore()
    for label_prf in f_per_type.values():
        micro_prf.tp += label_prf.tp
        micro_prf.fn += label_prf.fn
        micro_prf.fp += label_prf.fp
    n_labels = len(f_per_type) + 1e-100
    macro_p = sum(prf.precision for prf in f_per_type.values()) / n_labels
    macro_r = sum(prf.recall for prf in f_per_type.values()) / n_labels
    macro_f = sum(prf.fscore for prf in f_per_type.values()) / n_labels
    results = {
        f"nel_score": micro_prf.fscore,
        f"nel_score_desc": "micro F",
        f"nel_micro_p": micro_prf.precision,
        f"nel_micro_r": micro_prf.recall,
        f"nel_micro_f": micro_prf.fscore,
        f"nel_macro_p": macro_p,
        f"nel_macro_r": macro_r,
        f"nel_macro_f": macro_f,
        f"nel_f_per_type": {k: v.to_dict() for k, v in f_per_type.items()},
    }
    return results
